@@ -14,36 +14,54 @@
 
 # A tool to run tests in many different ways.
 
-import shlex
-import subprocess, sys, os, argparse
-import pickle
-from mesonbuild import build
-from mesonbuild import environment
-from mesonbuild.dependencies import ExternalProgram
-from mesonbuild import mesonlib
-from mesonbuild import mlog
-
-import time, datetime, multiprocessing, json
-import concurrent.futures as conc
-import platform
-import signal
-import random
+from collections import namedtuple
 from copy import deepcopy
+import argparse
+import concurrent.futures as conc
+import datetime
 import enum
+import io
+import json
+import multiprocessing
+import os
+import pickle
+import platform
+import random
+import re
+import shlex
+import signal
+import subprocess
+import sys
+import tempfile
+import time
+import typing
+
+from . import build
+from . import environment
+from . import mlog
+from .dependencies import ExternalProgram
+from .mesonlib import substring_is_in_list, MesonException
+
+if typing.TYPE_CHECKING:
+    from .backend.backends import TestSerialisation
 
 # GNU autotools interprets a return code of 77 from tests it executes to
 # mean that the test should be skipped.
 GNU_SKIP_RETURNCODE = 77
 
-def is_windows():
+# GNU autotools interprets a return code of 99 from tests it executes to
+# mean that the test failed even before testing what it is supposed to test.
+GNU_ERROR_RETURNCODE = 99
+
+def is_windows() -> bool:
     platname = platform.system().lower()
     return platname == 'windows' or 'mingw' in platname
 
-def is_cygwin():
+def is_cygwin() -> bool:
     platname = platform.system().lower()
     return 'cygwin' in platname
 
-def determine_worker_count():
+def determine_worker_count() -> int:
     varname = 'MESON_TESTTHREADS'
     if varname in os.environ:
         try:
@@ -60,8 +78,7 @@ def determine_worker_count():
             num_workers = 1
     return num_workers
 
-def buildparser():
-    parser = argparse.ArgumentParser(prog='meson test')
+def add_arguments(parser: argparse.ArgumentParser) -> None:
     parser.add_argument('--repeat', default=1, dest='repeat', type=int,
                         help='Number of times to run the tests.')
     parser.add_argument('--no-rebuild', default=False, action='store_true',
@@ -102,10 +119,9 @@ def buildparser():
                         help='Arguments to pass to the specified test(s) or all tests')
     parser.add_argument('args', nargs='*',
                         help='Optional list of tests to run')
-    return parser
 
 
-def returncode_to_status(retcode):
+def returncode_to_status(retcode: int) -> str:
     # Note: We can't use `os.WIFSIGNALED(result.returncode)` and the related
     # functions here because the status returned by subprocess is munged. It
     # returns a negative value if the process was killed by a signal rather than
@@ -130,8 +146,11 @@ def returncode_to_status(retcode):
         signame = 'SIGinvalid'
     return '(exit status %d or signal %d %s)' % (retcode, signum, signame)
 
+def env_tuple_to_str(env: typing.Iterable[typing.Tuple[str, str]]) -> str:
+    return ''.join(["%s='%s' " % (k, v) for k, v in env])
 
-class TestException(mesonlib.MesonException):
+
+class TestException(MesonException):
     pass
 
 
@@ -142,11 +161,215 @@ class TestResult(enum.Enum):
     TIMEOUT = 'TIMEOUT'
     SKIP = 'SKIP'
     FAIL = 'FAIL'
+    EXPECTEDFAIL = 'EXPECTEDFAIL'
+    UNEXPECTEDPASS = 'UNEXPECTEDPASS'
+    ERROR = 'ERROR'
+
+
+class TAPParser:
+    Plan = namedtuple('Plan', ['count', 'late', 'skipped', 'explanation'])
+    Bailout = namedtuple('Bailout', ['message'])
+    Test = namedtuple('Test', ['number', 'name', 'result', 'explanation'])
+    Error = namedtuple('Error', ['message'])
+    Version = namedtuple('Version', ['version'])
+
+    _MAIN = 1
+    _AFTER_TEST = 2
+    _YAML = 3
+
+    _RE_BAILOUT = re.compile(r'Bail out!\s*(.*)')
+    _RE_DIRECTIVE = re.compile(r'(?:\s*\#\s*([Ss][Kk][Ii][Pp]\S*|[Tt][Oo][Dd][Oo])\b\s*(.*))?')
+    _RE_PLAN = re.compile(r'1\.\.([0-9]+)' + _RE_DIRECTIVE.pattern)
+    _RE_TEST = re.compile(r'((?:not )?ok)\s*(?:([0-9]+)\s*)?([^#]*)' + _RE_DIRECTIVE.pattern)
+    _RE_VERSION = re.compile(r'TAP version ([0-9]+)')
+    _RE_YAML_START = re.compile(r'(\s+)---.*')
+    _RE_YAML_END = re.compile(r'\s+\.\.\.\s*')
+
+    def __init__(self, io: typing.Iterator[str]):
+        self.io = io
+
+    def parse_test(self, ok: bool, num: int, name: str, directive: typing.Optional[str], explanation: typing.Optional[str]) -> \
+            typing.Generator[typing.Union['TAPParser.Test', 'TAPParser.Error'], None, None]:
+        name = name.strip()
+        explanation = explanation.strip() if explanation else None
+        if directive is not None:
+            directive = directive.upper()
+            if directive == 'SKIP':
+                if ok:
+                    yield self.Test(num, name, TestResult.SKIP, explanation)
+                    return
+            elif directive == 'TODO':
+                yield self.Test(num, name, TestResult.UNEXPECTEDPASS if ok else TestResult.EXPECTEDFAIL, explanation)
+                return
+            else:
+                yield self.Error('invalid directive "%s"' % (directive,))
+
+        yield self.Test(num, name, TestResult.OK if ok else TestResult.FAIL, explanation)
+
+    def parse(self) -> typing.Generator[typing.Union['TAPParser.Test', 'TAPParser.Error', 'TAPParser.Version', 'TAPParser.Plan', 'TAPParser.Bailout'], None, None]:
+        found_late_test = False
+        bailed_out = False
+        plan = None
+        lineno = 0
+        num_tests = 0
+        yaml_lineno = None
+        yaml_indent = ''
+        state = self._MAIN
+        version = 12
+        while True:
+            lineno += 1
+            try:
+                line = next(self.io).rstrip()
+            except StopIteration:
+                break
+
+            # YAML blocks are only accepted after a test
+            if state == self._AFTER_TEST:
+                if version >= 13:
+                    m = self._RE_YAML_START.match(line)
+                    if m:
+                        state = self._YAML
+                        yaml_lineno = lineno
+                        yaml_indent = m.group(1)
+                        continue
+                state = self._MAIN
+
+            elif state == self._YAML:
+                if self._RE_YAML_END.match(line):
+                    state = self._MAIN
+                    continue
+                if line.startswith(yaml_indent):
+                    continue
+                yield self.Error('YAML block not terminated (started on line {})'.format(yaml_lineno))
+                state = self._MAIN
+
+            assert state == self._MAIN
+            if line.startswith('#'):
+                continue
+
+            m = self._RE_TEST.match(line)
+            if m:
+                if plan and plan.late and not found_late_test:
+                    yield self.Error('unexpected test after late plan')
+                    found_late_test = True
+                num_tests += 1
+                num = num_tests if m.group(2) is None else int(m.group(2))
+                if num != num_tests:
+                    yield self.Error('out of order test numbers')
+                yield from self.parse_test(m.group(1) == 'ok', num,
+                                           m.group(3), m.group(4), m.group(5))
+                state = self._AFTER_TEST
+                continue
+
+            m = self._RE_PLAN.match(line)
+            if m:
+                if plan:
+                    yield self.Error('more than one plan found')
+                else:
+                    count = int(m.group(1))
+                    skipped = (count == 0)
+                    if m.group(2):
+                        if m.group(2).upper().startswith('SKIP'):
+                            if count > 0:
+                                yield self.Error('invalid SKIP directive for plan')
+                            skipped = True
+                        else:
+                            yield self.Error('invalid directive for plan')
+                    plan = self.Plan(count=count, late=(num_tests > 0),
+                                     skipped=skipped, explanation=m.group(3))
+                    yield plan
+                continue
+
+            m = self._RE_BAILOUT.match(line)
+            if m:
+                yield self.Bailout(m.group(1))
+                bailed_out = True
+                continue
+
+            m = self._RE_VERSION.match(line)
+            if m:
+                # The TAP version is only accepted as the first line
+                if lineno != 1:
+                    yield self.Error('version number must be on the first line')
+                    continue
+                version = int(m.group(1))
+                if version < 13:
+                    yield self.Error('version number should be at least 13')
+                else:
+                    yield self.Version(version=version)
+                continue
+
+            yield self.Error('unexpected input at line %d' % (lineno,))
+
+        if state == self._YAML:
+            yield self.Error('YAML block not terminated (started on line {})'.format(yaml_lineno))
+
+        if not bailed_out and plan and num_tests != plan.count:
+            if num_tests < plan.count:
+                yield self.Error('Too few tests run (expected %d, got %d)' % (plan.count, num_tests))
+            else:
+                yield self.Error('Too many tests run (expected %d, got %d)' % (plan.count, num_tests))
 
 
 class TestRun:
-    def __init__(self, res, returncode, should_fail, duration, stdo, stde, cmd,
-                 env):
+
+    @classmethod
+    def make_exitcode(cls, test: 'TestSerialisation', test_env: typing.Dict[str, str],
+                      returncode: int, duration: float, stdo: typing.Optional[str],
+                      stde: typing.Optional[str],
+                      cmd: typing.Optional[typing.List[str]]) -> 'TestRun':
+        if returncode == GNU_SKIP_RETURNCODE:
+            res = TestResult.SKIP
+        elif returncode == GNU_ERROR_RETURNCODE:
+            res = TestResult.ERROR
+        elif test.should_fail:
+            res = TestResult.EXPECTEDFAIL if bool(returncode) else TestResult.UNEXPECTEDPASS
+        else:
+            res = TestResult.FAIL if bool(returncode) else TestResult.OK
+        return cls(test, test_env, res, returncode, duration, stdo, stde, cmd)
+
+    @classmethod
+    def make_tap(cls, test: 'TestSerialisation', test_env: typing.Dict[str, str],
+                 returncode: int, duration: float, stdo: str, stde: str,
+                 cmd: typing.Optional[typing.List[str]]) -> 'TestRun':
+        res = None
+        num_tests = 0
+        failed = False
+        num_skipped = 0
+
+        for i in TAPParser(io.StringIO(stdo)).parse():
+            if isinstance(i, TAPParser.Bailout):
+                res = TestResult.ERROR
+            elif isinstance(i, TAPParser.Test):
+                if i.result == TestResult.SKIP:
+                    num_skipped += 1
+                elif i.result in (TestResult.FAIL, TestResult.UNEXPECTEDPASS):
+                    failed = True
+                num_tests += 1
+            elif isinstance(i, TAPParser.Error):
+                res = TestResult.ERROR
+                stde += '\nTAP parsing error: ' + i.message
+
+        if returncode != 0:
+            res = TestResult.ERROR
+            stde += '\n(test program exited with status code %d)' % (returncode,)
+
+        if res is None:
+            # Now determine the overall result of the test based on the outcome of the subcases
+            if num_skipped == num_tests:
+                # This includes the case where num_tests is zero
+                res = TestResult.SKIP
+            elif test.should_fail:
+                res = TestResult.EXPECTEDFAIL if failed else TestResult.UNEXPECTEDPASS
+            else:
+                res = TestResult.FAIL if failed else TestResult.OK
+
+        return cls(test, test_env, res, returncode, duration, stdo, stde, cmd)
+
+    def __init__(self, test: 'TestSerialisation', test_env: typing.Dict[str, str],
+                 res: TestResult, returncode: int, duration: float,
+                 stdo: typing.Optional[str], stde: typing.Optional[str],
+                 cmd: typing.Optional[typing.List[str]]):
         assert isinstance(res, TestResult)
         self.res = res
         self.returncode = returncode
@@ -154,15 +377,16 @@ class TestRun:
         self.stdo = stdo
         self.stde = stde
         self.cmd = cmd
-        self.env = env
-        self.should_fail = should_fail
+        self.env = test_env
+        self.should_fail = test.should_fail
 
-    def get_log(self):
+    def get_log(self) -> str:
         res = '--- command ---\n'
         if self.cmd is None:
             res += 'NONE\n'
         else:
-            res += '%s%s\n' % (''.join(["%s='%s' " % (k, v) for k, v in self.env.items()]), ' ' .join(self.cmd))
+            test_only_env = set(self.env.items()) - set(os.environ.items())
+            res += '{}{}\n'.format(env_tuple_to_str(test_only_env), ' '.join(self.cmd))
         if self.stdo:
             res += '--- stdout ---\n'
             res += self.stdo
@@ -176,7 +400,7 @@ class TestRun:
         res += '-------\n\n'
         return res
 
-def decode(stream):
+def decode(stream: typing.Union[None, bytes]) -> str:
     if stream is None:
         return ''
     try:
@@ -184,51 +408,50 @@ def decode(stream):
     except UnicodeDecodeError:
         return stream.decode('iso-8859-1', errors='ignore')
 
-def write_json_log(jsonlogfile, test_name, result):
+def write_json_log(jsonlogfile: typing.TextIO, test_name: str, result: TestRun) -> None:
     jresult = {'name': test_name,
                'stdout': result.stdo,
                'result': result.res.value,
                'duration': result.duration,
                'returncode': result.returncode,
-               'command': result.cmd}
-    if isinstance(result.env, dict):
-        jresult['env'] = result.env
-    else:
-        jresult['env'] = result.env.get_env(os.environ)
+               'env': result.env,
+               'command': result.cmd}  # type: typing.Dict[str, typing.Any]
     if result.stde:
         jresult['stderr'] = result.stde
     jsonlogfile.write(json.dumps(jresult) + '\n')
 
-def run_with_mono(fname):
+def run_with_mono(fname: str) -> bool:
     if fname.endswith('.exe') and not (is_windows() or is_cygwin()):
         return True
     return False
 
-def load_benchmarks(build_dir):
+def load_benchmarks(build_dir: str) -> typing.List['TestSerialisation']:
     datafile = os.path.join(build_dir, 'meson-private', 'meson_benchmark_setup.dat')
     if not os.path.isfile(datafile):
         raise TestException('Directory ${!r} does not seem to be a Meson build directory.'.format(build_dir))
     with open(datafile, 'rb') as f:
-        obj = pickle.load(f)
+        obj = typing.cast(typing.List['TestSerialisation'], pickle.load(f))
     return obj
 
-def load_tests(build_dir):
+def load_tests(build_dir: str) -> typing.List['TestSerialisation']:
     datafile = os.path.join(build_dir, 'meson-private', 'meson_test_setup.dat')
     if not os.path.isfile(datafile):
         raise TestException('Directory ${!r} does not seem to be a Meson build directory.'.format(build_dir))
     with open(datafile, 'rb') as f:
-        obj = pickle.load(f)
+        obj = typing.cast(typing.List['TestSerialisation'], pickle.load(f))
     return obj
 
 
 class SingleTestRunner:
 
-    def __init__(self, test, env, options):
+    def __init__(self, test: 'TestSerialisation', test_env: typing.Dict[str, str],
+                 env: typing.Dict[str, str], options: argparse.Namespace):
         self.test = test
+        self.test_env = test_env
         self.env = env
         self.options = options
 
-    def _get_cmd(self):
+    def _get_cmd(self) -> typing.Optional[typing.List[str]]:
         if self.test.fname[0].endswith('.jar'):
             return ['java', '-jar'] + self.test.fname
         elif not self.test.is_cross_built and run_with_mono(self.test.fname[0]):
@@ -240,36 +463,31 @@ class SingleTestRunner:
                     # because there is no execute wrapper.
                     return None
                 else:
-                    return [self.test.exe_runner] + self.test.fname
+                    if not self.test.exe_runner.found():
+                        msg = 'The exe_wrapper defined in the cross file {!r} was not ' \
+                              'found. Please check the command and/or add it to PATH.'
+                        raise TestException(msg.format(self.test.exe_runner.name))
+                    return self.test.exe_runner.get_command() + self.test.fname
             else:
                 return self.test.fname
 
-    def run(self):
+    def run(self) -> TestRun:
         cmd = self._get_cmd()
         if cmd is None:
             skip_stdout = 'Not run because can not execute cross compiled binaries.'
-            return TestRun(res=TestResult.SKIP, returncode=GNU_SKIP_RETURNCODE,
-                           should_fail=self.test.should_fail, duration=0.0,
-                           stdo=skip_stdout, stde=None, cmd=None, env=self.test.env)
+            return TestRun(self.test, self.test_env, TestResult.SKIP, GNU_SKIP_RETURNCODE, 0.0, skip_stdout, None, None)
         else:
             wrap = TestHarness.get_wrapper(self.options)
             if self.options.gdb:
                 self.test.timeout = None
             return self._run_cmd(wrap + cmd + self.test.cmd_args + self.options.test_args)
 
-    @staticmethod
-    def _substring_in_list(substr, strlist):
-        for s in strlist:
-            if substr in s:
-                return True
-        return False
-
-    def _run_cmd(self, cmd):
+    def _run_cmd(self, cmd: typing.List[str]) -> TestRun:
         starttime = time.time()
 
         if len(self.test.extra_paths) > 0:
             self.env['PATH'] = os.pathsep.join(self.test.extra_paths + ['']) + self.env['PATH']
-            if self._substring_in_list('wine', cmd):
+            if substring_is_in_list('wine', cmd):
                 wine_paths = ['Z:' + p for p in self.test.extra_paths]
                 wine_path = ';'.join(wine_paths)
                 # Don't accidentally end with an `;` because that will add the
@@ -290,8 +508,10 @@ class SingleTestRunner:
         stdout = None
         stderr = None
         if not self.options.verbose:
-            stdout = subprocess.PIPE
-            stderr = subprocess.PIPE if self.options and self.options.split else subprocess.STDOUT
+            stdout = tempfile.TemporaryFile("wb+")
+            stderr = tempfile.TemporaryFile("wb+") if self.options.split else stdout
+        if self.test.protocol == 'tap' and stderr is stdout:
+            stdout = tempfile.TemporaryFile("wb+")
 
         # Let gdb handle ^C instead of us
         if self.options.gdb:
@@ -299,7 +519,7 @@ class SingleTestRunner:
             # Make the meson executable ignore SIGINT while gdb is running.
             signal.signal(signal.SIGINT, signal.SIG_IGN)
 
-        def preexec_fn():
+        def preexec_fn() -> None:
             if self.options.gdb:
                 # Restore the SIGINT handler for the child process to
                 # ensure it can handle it.
@@ -325,10 +545,10 @@ class SingleTestRunner:
         else:
             timeout = self.test.timeout
         try:
-            (stdo, stde) = p.communicate(timeout=timeout)
+            p.communicate(timeout=timeout)
         except subprocess.TimeoutExpired:
             if self.options.verbose:
-                print('%s time out (After %d seconds)' % (self.test.name, timeout))
+                print('{} time out (After {} seconds)'.format(self.test.name, timeout))
             timed_out = True
         except KeyboardInterrupt:
             mlog.warning('CTRL-C detected while running %s' % (self.test.name))
@@ -338,6 +558,8 @@ class SingleTestRunner:
                 # Let us accept ^C again
                 signal.signal(signal.SIGINT, previous_sigint_handler)
 
+        additional_error = None
+
         if kill_test or timed_out:
             # Python does not provide multiplatform support for
             # killing a process and all its children so we need
@@ -353,50 +575,78 @@ class SingleTestRunner:
                     # There's nothing we can do (maybe the process
                     # already died) so carry on.
                     pass
-            (stdo, stde) = p.communicate()
+            try:
+                p.communicate(timeout=1)
+            except subprocess.TimeoutExpired:
+                # An earlier kill attempt has not worked for whatever reason.
+                # Try to kill it one last time with a direct call.
+                # If the process has spawned children, they will remain around.
+                p.kill()
+                try:
+                    p.communicate(timeout=1)
+                except subprocess.TimeoutExpired:
+                    additional_error = 'Test process could not be killed.'
+            except ValueError:
+                additional_error = 'Could not read output. Maybe the process has redirected its stdout/stderr?'
         endtime = time.time()
         duration = endtime - starttime
-        stdo = decode(stdo)
-        if stde:
-            stde = decode(stde)
-        if timed_out:
-            res = TestResult.TIMEOUT
-        elif p.returncode == GNU_SKIP_RETURNCODE:
-            res = TestResult.SKIP
-        elif self.test.should_fail == bool(p.returncode):
-            res = TestResult.OK
+        if additional_error is None:
+            if stdout is None:
+                stdo = ''
+            else:
+                stdout.seek(0)
+                stdo = decode(stdout.read())
+            if stderr is None or stderr is stdout:
+                stde = ''
+            else:
+                stderr.seek(0)
+                stde = decode(stderr.read())
         else:
-            res = TestResult.FAIL
-        return TestRun(res, p.returncode, self.test.should_fail, duration, stdo, stde, cmd, self.test.env)
+            stdo = ""
+            stde = additional_error
+        if timed_out:
+            return TestRun(self.test, self.test_env, TestResult.TIMEOUT, p.returncode, duration, stdo, stde, cmd)
+        else:
+            if self.test.protocol == 'exitcode':
+                return TestRun.make_exitcode(self.test, self.test_env, p.returncode, duration, stdo, stde, cmd)
+            else:
+                if self.options.verbose:
+                    print(stdo, end='')
+                return TestRun.make_tap(self.test, self.test_env, p.returncode, duration, stdo, stde, cmd)
 
 
 class TestHarness:
-    def __init__(self, options):
+    def __init__(self, options: argparse.Namespace):
         self.options = options
-        self.collected_logs = []
+        self.collected_logs = []  # type: typing.List[str]
         self.fail_count = 0
+        self.expectedfail_count = 0
+        self.unexpectedpass_count = 0
         self.success_count = 0
         self.skip_count = 0
         self.timeout_count = 0
         self.is_run = False
         self.tests = None
-        self.suites = None
-        self.logfilename = None
-        self.logfile = None
-        self.jsonlogfile = None
+        self.logfilename = None   # type: typing.Optional[str]
+        self.logfile = None       # type: typing.Optional[typing.TextIO]
+        self.jsonlogfile = None   # type: typing.Optional[typing.TextIO]
         if self.options.benchmark:
             self.tests = load_benchmarks(options.wd)
         else:
             self.tests = load_tests(options.wd)
-        self.load_suites()
+        ss = set()
+        for t in self.tests:
+            for s in t.suite:
+                ss.add(s)
+        self.suites = list(ss)
 
-    def __del__(self):
+    def __del__(self) -> None:
         if self.logfile:
             self.logfile.close()
         if self.jsonlogfile:
             self.jsonlogfile.close()
 
-    def merge_suite_options(self, options, test):
+    def merge_suite_options(self, options: argparse.Namespace, test: 'TestSerialisation') -> typing.Dict[str, str]:
         if ':' in options.setup:
             if options.setup not in self.build_data.test_setups:
                 sys.exit("Unknown test setup '%s'." % options.setup)
@@ -408,6 +658,8 @@ class TestHarness:
             current = self.build_data.test_setups[full_name]
         if not options.gdb:
             options.gdb = current.gdb
+        if options.gdb:
+            options.verbose = True
         if options.timeout_multiplier is None:
             options.timeout_multiplier = current.timeout_multiplier
     #    if options.env is None:
@@ -418,31 +670,36 @@ class TestHarness:
             options.wrapper = current.exe_wrapper
         return current.env.get_env(os.environ.copy())
 
-    def get_test_runner(self, test):
+    def get_test_runner(self, test: 'TestSerialisation') -> SingleTestRunner:
         options = deepcopy(self.options)
+        if not options.setup:
+            options.setup = self.build_data.test_setup_default_name
         if options.setup:
             env = self.merge_suite_options(options, test)
         else:
             env = os.environ.copy()
-        if isinstance(test.env, build.EnvironmentVariables):
-            test.env = test.env.get_env(env)
-        env.update(test.env)
-        return SingleTestRunner(test, env, options)
+        test_env = test.env.get_env(env)
+        env.update(test_env)
+        return SingleTestRunner(test, test_env, env, options)
 
-    def process_test_result(self, result):
+    def process_test_result(self, result: TestRun) -> None:
         if result.res is TestResult.TIMEOUT:
             self.timeout_count += 1
-            self.fail_count += 1
         elif result.res is TestResult.SKIP:
             self.skip_count += 1
         elif result.res is TestResult.OK:
             self.success_count += 1
-        elif result.res is TestResult.FAIL:
+        elif result.res is TestResult.FAIL or result.res is TestResult.ERROR:
             self.fail_count += 1
+        elif result.res is TestResult.EXPECTEDFAIL:
+            self.expectedfail_count += 1
+        elif result.res is TestResult.UNEXPECTEDPASS:
+            self.unexpectedpass_count += 1
         else:
             sys.exit('Unknown test result encountered: {}'.format(result.res))
 
-    def print_stats(self, numlen, tests, name, result, i):
+    def print_stats(self, numlen: int, tests: typing.List['TestSerialisation'],
+                    name: str, result: TestRun, i: int) -> None:
         startpad = ' ' * (numlen - len('%d' % (i + 1)))
         num = '%s%d/%d' % (startpad, i + 1, len(tests))
         padding1 = ' ' * (38 - len(name))
@@ -454,9 +711,12 @@ class TestHarness:
         result_str = '%s %s  %s%s%s%5.2f s %s' % \
             (num, name, padding1, result.res.value, padding2, result.duration,
              status)
-        if not self.options.quiet or result.res is not TestResult.OK:
-            if result.res is not TestResult.OK and mlog.colorize_console:
-                if result.res in (TestResult.FAIL, TestResult.TIMEOUT):
+        ok_statuses = (TestResult.OK, TestResult.EXPECTEDFAIL)
+        bad_statuses = (TestResult.FAIL, TestResult.TIMEOUT, TestResult.UNEXPECTEDPASS,
+                        TestResult.ERROR)
+        if not self.options.quiet or result.res not in ok_statuses:
+            if result.res not in ok_statuses and mlog.colorize_console:
+                if result.res in bad_statuses:
                     decorator = mlog.red
                 elif result.res is TestResult.SKIP:
                     decorator = mlog.yellow
@@ -466,8 +726,7 @@ class TestHarness:
             else:
                 print(result_str)
         result_str += "\n\n" + result.get_log()
-        if (result.returncode != GNU_SKIP_RETURNCODE) \
-                and (result.returncode != 0) != result.should_fail:
+        if result.res in bad_statuses:
             if self.options.print_errorlogs:
                 self.collected_logs.append(result_str)
         if self.logfile:
@@ -475,18 +734,21 @@ class TestHarness:
         if self.jsonlogfile:
             write_json_log(self.jsonlogfile, name, result)
 
-    def print_summary(self):
+    def print_summary(self) -> None:
         msg = '''
-OK:      %4d
-FAIL:    %4d
-SKIP:    %4d
-TIMEOUT: %4d
-''' % (self.success_count, self.fail_count, self.skip_count, self.timeout_count)
+Ok:                 %4d
+Expected Fail:      %4d
+Fail:               %4d
+Unexpected Pass:    %4d
+Skipped:            %4d
+Timeout:            %4d
+''' % (self.success_count, self.expectedfail_count, self.fail_count,
+            self.unexpectedpass_count, self.skip_count, self.timeout_count)
         print(msg)
         if self.logfile:
             self.logfile.write(msg)
 
-    def print_collected_logs(self):
+    def print_collected_logs(self) -> None:
         if len(self.collected_logs) > 0:
             if len(self.collected_logs) > 10:
                 print('\nThe output from 10 first failed tests:\n')
@@ -499,9 +761,16 @@ TIMEOUT: %4d
                     print('--- Listing only the last 100 lines from a long log. ---')
                     lines = lines[-100:]
                 for line in lines:
-                    print(line)
+                    try:
+                        print(line)
+                    except UnicodeEncodeError:
+                        line = line.encode('ascii', errors='replace').decode()
+                        print(line)
 
-    def doit(self):
+    def total_failure_count(self) -> int:
+        return self.fail_count + self.unexpectedpass_count + self.timeout_count
+
+    def doit(self) -> int:
         if self.is_run:
             raise RuntimeError('Test harness object can only be used once.')
         self.is_run = True
@@ -509,17 +778,19 @@ TIMEOUT: %4d
         if not tests:
             return 0
         self.run_tests(tests)
-        return self.fail_count
+        return self.total_failure_count()
 
     @staticmethod
-    def split_suite_string(suite):
+    def split_suite_string(suite: str) -> typing.Tuple[str, str]:
         if ':' in suite:
-            return suite.split(':', 1)
+            # mypy can't figure out that str.split(n, 1) will return a list of
+            # length 2, so we have to help it.
+            return typing.cast(typing.Tuple[str, str], tuple(suite.split(':', 1)))
         else:
             return suite, ""
 
     @staticmethod
-    def test_in_suites(test, suites):
+    def test_in_suites(test: 'TestSerialisation', suites: typing.List[str]) -> bool:
         for suite in suites:
             (prj_match, st_match) = TestHarness.split_suite_string(suite)
             for prjst in test.suite:
@@ -550,18 +821,11 @@ TIMEOUT: %4d
                 return True
         return False
 
-    def test_suitable(self, test):
+    def test_suitable(self, test: 'TestSerialisation') -> bool:
         return (not self.options.include_suites or TestHarness.test_in_suites(test, self.options.include_suites)) \
             and not TestHarness.test_in_suites(test, self.options.exclude_suites)
 
-    def load_suites(self):
-        ss = set()
-        for t in self.tests:
-            for s in t.suite:
-                ss.add(s)
-        self.suites = list(ss)
-
-    def get_tests(self):
+    def get_tests(self) -> typing.List['TestSerialisation']:
         if not self.tests:
             print('No tests defined.')
             return []
@@ -581,14 +845,11 @@ TIMEOUT: %4d
             print('No suitable tests defined.')
             return []
 
-        for test in tests:
-            test.rebuilt = False
-
         return tests
 
-    def open_log_files(self):
+    def open_log_files(self) -> None:
         if not self.options.logbase or self.options.verbose:
-            return None, None, None, None
+            return
 
         namebase = None
         logfile_base = os.path.join(self.options.wd, 'meson-logs', self.options.logbase)
@@ -603,15 +864,17 @@ TIMEOUT: %4d
         self.logfilename = logfile_base + '.txt'
         self.jsonlogfilename = logfile_base + '.json'
 
-        self.jsonlogfile = open(self.jsonlogfilename, 'w', encoding='utf-8')
-        self.logfile = open(self.logfilename, 'w', encoding='utf-8')
+        self.jsonlogfile = open(self.jsonlogfilename, 'w', encoding='utf-8', errors='replace')
+        self.logfile = open(self.logfilename, 'w', encoding='utf-8', errors='surrogateescape')
 
         self.logfile.write('Log of Meson test suite run on %s\n\n'
                            % datetime.datetime.now().isoformat())
+        inherit_env = env_tuple_to_str(os.environ.items())
+        self.logfile.write('Inherited environment: {}\n\n'.format(inherit_env))
 
     @staticmethod
-    def get_wrapper(options):
-        wrap = []
+    def get_wrapper(options: argparse.Namespace) -> typing.List[str]:
+        wrap = []  # type: typing.List[str]
         if options.gdb:
             wrap = ['gdb', '--quiet', '--nh']
             if options.repeat > 1:
@@ -620,11 +883,10 @@ TIMEOUT: %4d
             wrap += ['--args']
         if options.wrapper:
             wrap += options.wrapper
-        assert(isinstance(wrap, list))
         return wrap
 
-    def get_pretty_suite(self, test):
-        if len(self.suites) > 1:
+    def get_pretty_suite(self, test: 'TestSerialisation') -> str:
+        if len(self.suites) > 1 and test.suite:
             rv = TestHarness.split_suite_string(test.suite[0])[0]
             s = "+".join(TestHarness.split_suite_string(s)[1] for s in test.suite)
             if len(s):
@@ -633,9 +895,9 @@ TIMEOUT: %4d
         else:
             return test.name
 
-    def run_tests(self, tests):
+    def run_tests(self, tests: typing.List['TestSerialisation']) -> None:
         executor = None
-        futures = []
+        futures = []  # type: typing.List[typing.Tuple[conc.Future[TestRun], int, typing.List[TestSerialisation], str, int]]
         numlen = len('%d' % len(tests))
         self.open_log_files()
         startdir = os.getcwd()
@@ -647,18 +909,17 @@ TIMEOUT: %4d
             for _ in range(self.options.repeat):
                 for i, test in enumerate(tests):
                     visible_name = self.get_pretty_suite(test)
+                    single_test = self.get_test_runner(test)
 
-                    if not test.is_parallel or self.options.gdb:
+                    if not test.is_parallel or self.options.num_processes == 1 or single_test.options.gdb:
                         self.drain_futures(futures)
                         futures = []
-                        single_test = self.get_test_runner(test)
                         res = single_test.run()
                         self.process_test_result(res)
                         self.print_stats(numlen, tests, visible_name, res, i)
                     else:
                         if not executor:
                             executor = conc.ThreadPoolExecutor(max_workers=self.options.num_processes)
-                        single_test = self.get_test_runner(test)
                         f = executor.submit(single_test.run)
                         futures.append((f, numlen, tests, visible_name, i))
                     if self.options.repeat > 1 and self.fail_count:
@@ -675,9 +936,9 @@ TIMEOUT: %4d
         finally:
             os.chdir(startdir)
 
-    def drain_futures(self, futures):
-        for i in futures:
-            (result, numlen, tests, name, i) = i
+    def drain_futures(self, futures: typing.List[typing.Tuple['conc.Future[TestRun]', int, typing.List['TestSerialisation'], str, int]]) -> None:
+        for x in futures:
+            (result, numlen, tests, name, i) = x
             if self.options.repeat > 1 and self.fail_count:
                 result.cancel()
             if self.options.verbose:
@@ -685,7 +946,7 @@ TIMEOUT: %4d
             self.process_test_result(result.result())
             self.print_stats(numlen, tests, name, result.result(), i)
 
-    def run_special(self):
+    def run_special(self) -> int:
         '''Tests run by the user, usually something like "under gdb 1000 times".'''
         if self.is_run:
             raise RuntimeError('Can not use run_special after a full run.')
@@ -693,15 +954,16 @@ TIMEOUT: %4d
         if not tests:
             return 0
         self.run_tests(tests)
-        return self.fail_count
+        return self.total_failure_count()
 
 
-def list_tests(th):
+def list_tests(th: TestHarness) -> bool:
     tests = th.get_tests()
     for t in tests:
         print(th.get_pretty_suite(t))
+    return not tests
 
-def rebuild_all(wd):
+def rebuild_all(wd: str) -> bool:
     if not os.path.isfile(os.path.join(wd, 'build.ninja')):
         print('Only ninja backend is supported to rebuild tests before running them.')
         return True
@@ -720,9 +982,7 @@ def rebuild_all(wd):
 
     return True
 
-def run(args):
-    options = buildparser().parse_args(args)
-
+def run(options: argparse.Namespace) -> int:
     if options.benchmark:
         options.num_processes = 1
 
@@ -744,22 +1004,31 @@ def run(args):
     if check_bin is not None:
         exe = ExternalProgram(check_bin, silent=True)
         if not exe.found():
-            sys.exit('Could not find requested program: %s' % check_bin)
+            print('Could not find requested program: {!r}'.format(check_bin))
+            return 1
     options.wd = os.path.abspath(options.wd)
 
     if not options.list and not options.no_rebuild:
         if not rebuild_all(options.wd):
-            sys.exit(-1)
+            return 1
 
     try:
         th = TestHarness(options)
         if options.list:
-            list_tests(th)
-            return 0
+            return list_tests(th)
         if not options.args:
             return th.doit()
         return th.run_special()
     except TestException as e:
         print('Meson test encountered an error:\n')
-        print(e)
+        if os.environ.get('MESON_FORCE_BACKTRACE'):
+            raise e
+        else:
+            print(e)
         return 1
+
+def run_with_args(args: typing.List[str]) -> int:
+    parser = argparse.ArgumentParser(prog='meson test')
+    add_arguments(parser)
+    options = parser.parse_args(args)
+    return run(options)
