@@ -136,7 +136,7 @@ def find_vcxproj_with_target(builddir, target):
         p = r'<TargetName>{}</TargetName>\s*<TargetExt>\{}</TargetExt>'.format(t, ext)
     else:
         p = r'<TargetName>{}</TargetName>'.format(t)
-    for root, dirs, files in os.walk(builddir):
+    for _, _, files in os.walk(builddir):
         for f in fnmatch.filter(files, '*.vcxproj'):
             f = os.path.join(builddir, f)
             with open(f, 'r', encoding='utf-8') as o:
@@ -220,7 +220,7 @@ def clear_meson_configure_class_caches():
     mesonbuild.compilers.CCompiler.find_library_cache = {}
     mesonbuild.compilers.CCompiler.find_framework_cache = {}
     mesonbuild.dependencies.PkgConfigDependency.pkgbin_cache = {}
-    mesonbuild.dependencies.PkgConfigDependency.class_pkgbin = mesonlib.PerMachine(None, None, None)
+    mesonbuild.dependencies.PkgConfigDependency.class_pkgbin = mesonlib.PerMachine(None, None)
 
 def run_configure_inprocess(commandlist):
     old_stdout = sys.stdout
@@ -292,7 +292,7 @@ def main():
         os.environ.pop('platform')
     # Run tests
     print(mlog.bold('Running unittests.').get_text(mlog.colorize_console))
-    print()
+    print(flush=True)
     # Can't pass arguments to unit tests, so set the backend to use in the environment
     env = os.environ.copy()
     env['MESON_UNIT_TEST_BACKEND'] = backend.name
@@ -325,7 +325,7 @@ def main():
         else:
             cross_test_args = mesonlib.python_command + ['run_cross_test.py']
             print(mlog.bold('Running armhf cross tests.').get_text(mlog.colorize_console))
-            print()
+            print(flush=True)
             cmd = cross_test_args + ['cross/ubuntu-armhf.txt']
             if options.failfast:
                 cmd += ['--failfast']
@@ -334,7 +334,7 @@ def main():
                 return returncode
             print(mlog.bold('Running mingw-w64 64-bit cross tests.')
                   .get_text(mlog.colorize_console))
-            print()
+            print(flush=True)
             cmd = cross_test_args + ['cross/linux-mingw-w64-64bit.txt']
             if options.failfast:
                 cmd += ['--failfast']
