@@ -1,4 +1,4 @@
-submodule (mother) daughter
+submodule (parent) parent
 
 contains
 
@@ -6,5 +6,5 @@ module procedure pi2tau
   pi2tau = 2*pi
 end procedure pi2tau
 
-end submodule daughter
+end submodule parent
 
